import ujson
import functools

from .configuration import settings


def action(*verbs, encoding='utf-8', content_type=None, inner_decorator=None):
    def _decorator(func):
        argcount = func.__code__.co_argcount

        if inner_decorator is not None:
            func = inner_decorator(func)

        func.__http_methods__ = verbs if verbs else 'any'
        func.__response_encoding__ = encoding
        func.__argcount__ = argcount

        if content_type:
            func.__content_type__ = content_type

        return func

    if verbs and callable(verbs[0]):
        f = verbs[0]
        verbs = tuple()
        return _decorator(f)
    else:
        return _decorator


def jsonify(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if hasattr(result, 'to_dict'):
            result = result.to_dict()
        elif not isinstance(result, (list, dict, int, str)):
            raise ValueError('Cannot encode to json: %s' % type(result))

        return ujson.dumps(result, indent=settings.json.indent)

    return wrapper


html = functools.partial(action, content_type='text/html')
text = functools.partial(action, content_type='text/plain')
json = functools.partial(action, content_type='application/json', inner_decorator=jsonify)
xml = functools.partial(action, content_type='application/xml')
binary = functools.partial(action, content_type='application/octet-stream', encoding=None)




from os.path import isdir, join
import threading

import pymlconf

from .configuration import settings


class LazyAttribute(object):
    """ ``attribute`` decorator is intended to promote a
        function call to object attribute. This means the
        function is called once and replaced with
        returned value.

        >>> class A:
        ...     def __init__(self):
        ...         self.counter = 0
        ...     @LazyAttribute
        ...     def count(self):
        ...         self.counter += 1
        ...         return self.counter
        >>> a = A()
        >>> a.count
        1
        >>> a.count
        1
    """
    __slots__ = ('f', )

    def __init__(self, f):
        self.f = f

    def __get__(self, obj, t=None):
        f = self.f
        val = f(obj)
        setattr(obj, f.__name__, val)
        return val


def load_controller_from_file(specifier):
    from importlib.util import spec_from_file_location, module_from_spec
    controller = None

    if specifier:
        module_name, class_name = specifier.split(':') if ':' in specifier else (specifier, 'Root')

        if module_name:

            if isdir(module_name):
                location = join(module_name, '__init__.py')
            elif module_name.endswith('.py'):
                location = module_name
                module_name = module_name[:-3]
            else:
                location = '%s.py' % module_name

            spec = spec_from_file_location(module_name, location=location)
            module = module_from_spec(spec)
            spec.loader.exec_module(module)
            controller = getattr(module, class_name)()

        elif class_name == 'Static':
            from .controllers import Static
            controller = Static()
        else:  # pragma: no coverg
            controller = globals()[class_name]()

    return controller


def quickstart(controller=None, host='localhost',  port=8080, block=True, config=None):
    from wsgiref.simple_server import make_server

    try:
        settings.proxied_object
    except pymlconf.ConfigurationNotInitializedError:
        settings.load()

    if config:
        settings.merge(config)

    if controller is None:
        from wsgiref.simple_server import demo_app
        app = demo_app
    else:
        app = controller.load_app()

    port = int(port)
    httpd = make_server(host, port, app)

    print("Serving http://%s:%d" % (host or 'localhost', port))
    if block:  # pragma: no cover
        httpd.serve_forever()
    else:
        t = threading.Thread(target=httpd.serve_forever, daemon=True)
        t.start()

        def shutdown():
            httpd.shutdown()
            httpd.server_close()
            t.join()

        return shutdown




import re

from nanohttp import Controller, action, text, html, context
from nanohttp.tests.helpers import WsgiAppTestCase


class HttpHeadersTestCase(WsgiAppTestCase):

    class Root(Controller):

        @text()
        def index(self):
            yield 'Index'

        @action()
        def no_content_type(self):
            yield'No Content Type'

        @html()
        def html(self):
            yield context.response_content_type

        @action
        def custom_header(self):
            import time
            context.response_headers.add_header('my-header',  str(time.time()))
            yield ''

    def test_response_header(self):
        self.assert_get('/', expected_headers={'Content-Type': 'text/plain; charset=utf-8'})
        self.assert_get('/html', expected_headers={'Content-Type': 'text/html; charset=utf-8'})
        self.assert_get('/no_content_type', not_expected_headers=['Content-Type'])
        self.assert_get('/custom_header', expected_headers={
            'my-header': re. compile('\d+.?\d*')
        })



from nanohttp import Controller, action, text, HttpMovedPermanently, HttpFound
from nanohttp.tests.helpers import WsgiAppTestCase


class HttpRedirectTestCase(WsgiAppTestCase):

    class Root(Controller):

        @text()
        def index(self):
            raise HttpMovedPermanently('/new/address')

        @action()
        def about(self):
            raise HttpFound('/new/address')

    def test_redirect_response_header(self):
        self.assert_get('/', status=301, expected_headers={'Location': '/new/address'})
        self.assert_get('/about', status=302, expected_headers={'Location': '/new/address'})





import time
import os
import sys
import logging
import traceback
from os.path import isdir, join, relpath, pardir, exists
from mimetypes import guess_type

import ujson

from .configuration import settings
from .exceptions import HttpStatus, InternalServerError, HttpNotFound, HttpMethodNotAllowed, HttpForbidden
from .contexts import context, Context
from .constants import HTTP_DATETIME_FORMAT


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('nanohttp')


class Controller(object):
    __logger__ = logger
    __http_methods__ = 'any'
    __response_encoding__ = 'utf8'
    __default_action__ = 'index'
    __remove_trailing_slash__ = True

    def _hook(self, name, *args, **kwargs):
        if hasattr(self, name):
            return getattr(self, name)(*args, **kwargs)

    def load_app(self):
        self._hook('app_load')
        return self._handle_request

    def _handle_exception(self, ex):
        context.response_encoding = 'utf-8'

        error = ex if isinstance(ex, HttpStatus) else InternalServerError(sys.exc_info())
        error_page = self._hook('request_error', error)
        message = error.status_text
        description = error.render() if settings.debug else error.info if error_page is None else error_page

        if context.response_content_type == 'application/json':
            response = ujson.encode(dict(
                message=message,
                description=description
            ))
        else:
            context.response_content_type = 'text/plain'
            response = "%s\n%s" % (message, description)

        if isinstance(error, InternalServerError):
            traceback.print_exc()

        def resp():
            yield response

        return error.status, resp()

    def _handle_request(self, environ, start_response):
        ctx = Context(environ)
        ctx.__enter__()
        # start_response("200 OK", [('Content-Type', 'text/plain; charset=utf-8')])

        status = '200 OK'
        buffer = None

        try:
            self._hook('begin_request')
            if self.__remove_trailing_slash__:
                ctx.path = ctx.path.rstrip('/')

            result = self(*ctx.path.split('?')[0][1:].split('/'))
            if result:
                resp_generator = iter(result)
                buffer = next(resp_generator)
            else:
                resp_generator = None

        except Exception as ex:
            self.__logger__.exception('Exception while handling the request.')
            status, resp_generator = self._handle_exception(ex)

        finally:
            self._hook('begin_response')

            if context.response_cookies:
                for cookie in context.response_cookies:
                    ctx.response_headers.add_header(*cookie.http_set_cookie())
            start_response(status, ctx.response_headers.items())

        def _response():
            try:
                if buffer is not None:
                    yield ctx.encode_response(buffer)

                if resp_generator:
                    # noinspection PyTypeChecker
                    for chunk in resp_generator:
                        yield ctx.encode_response(chunk)
                else:
                    yield b''
            except Exception as ex_:  # pragma: no cover
                self.__logger__.exception('Exception while serving the response.')
                if settings.debug:
                    yield str(ex_).encode()
                raise ex_

            finally:
                self._hook('end_response')
                context.__exit__(*sys.exc_info())

        return _response()

    # noinspection PyMethodMayBeStatic
    def _serve_handler(self, handler, *remaining_paths):
        if hasattr(handler, '__response_encoding__'):
            context.response_encoding = handler.__response_encoding__

        if hasattr(handler, '__content_type__'):
            context.response_content_type = handler.__content_type__

        return handler(*remaining_paths)

    def _dispatch(self, *remaining_paths):
        if not len(remaining_paths):
            path = self.__default_action__
        else:
            path = self.__default_action__ if remaining_paths[0] == '' else remaining_paths[0]
            remaining_paths = remaining_paths[1:]

        # Ensuring the handler
        handler = getattr(self, path, None)
        if handler is None:
            handler = getattr(self, self.__default_action__, None)
            if handler is not None:
                remaining_paths = (path, ) + remaining_paths

        args_count = len(remaining_paths)
        if (handler is None or not hasattr(handler, '__http_methods__')) \
                or (hasattr(handler, '__argcount__') and handler.__argcount__ < args_count) \
                or (hasattr(handler, '__annotations__') and len(handler.__annotations__) < args_count):
            raise HttpNotFound()

        if 'any' != handler.__http_methods__ and context.method not in handler.__http_methods__:
            raise HttpMethodNotAllowed()

        return handler, remaining_paths

    def __call__(self, *remaining_paths):
        handler, remaining_paths = self._dispatch(*remaining_paths)
        return self._serve_handler(handler, *remaining_paths)


class RestController(Controller):
    __detect_verb_by_header__ = False

    def _dispatch(self, *remaining_paths):
        handler = None
        if remaining_paths:
            first_path = remaining_paths[0]
            if hasattr(self, first_path):
                remaining_paths = remaining_paths[1:]
                handler = getattr(self, first_path)

        if handler is None:
            if not hasattr(self, context.method):
                raise HttpMethodNotAllowed()

            if self.__detect_verb_by_header__ and context.environ.get(self.__detect_verb_by_header__):
                handler = getattr(self, context.environ.get(self.__detect_verb_by_header__))
            else:
                handler = getattr(self, context.method)

        if handler is None or not hasattr(handler, '__http_methods__'):
            raise HttpNotFound()

        # FIXME: check argcount
        if hasattr(handler, '__annotations__') and len(handler.__annotations__) < len(remaining_paths):
            raise HttpNotFound()

        return handler, remaining_paths


class Static(Controller):
    __response_encoding__ = None
    __chunk_size__ = 0x4000

    def __init__(self, directory='.', default_document='index.html'):
        self.default_document = default_document
        self.directory = directory

    def __call__(self, *remaining_paths):

        # Find the physical path of the given path parts
        physical_path = join(self.directory, *remaining_paths)

        # Check to do not access the parent directory of root and also we are not listing directories here.
        if pardir in relpath(physical_path, self.directory):
            raise HttpForbidden()

        if isdir(physical_path):
            if self.default_document:
                physical_path = join(physical_path, self.default_document)
                if not exists(physical_path):
                    raise HttpForbidden
            else:
                raise HttpForbidden()

        context.response_headers.add_header('Content-Type', guess_type(physical_path)[0] or 'application/octet-stream')

        try:
            f = open(physical_path, mode='rb')
            stat = os.fstat(f.fileno())
            context.response_headers.add_header('Content-Length', str(stat[6]))
            context.response_headers.add_header(
                'Last-Modified',
                time.strftime(HTTP_DATETIME_FORMAT, time.gmtime(stat.st_mtime))
            )

            with f:
                while True:
                    r = f.read(self.__chunk_size__)
                    if not r:
                        break
                    yield r

        except OSError:
            raise HttpNotFound()




import threading
import time
import unittest
from os.path import join, dirname, abspath

import httplib2

from nanohttp import main
from nanohttp.tests.helpers import find_free_tcp_port


class EntryPointTestCase(unittest.TestCase):

    def setUp(self):
        this_dir = abspath(dirname(__file__))
        self.demo_filename = join(this_dir, 'stuff/demo.txt')
        self.module_dir = abspath(join(this_dir, '..'))
        with open(self.demo_filename, mode='w') as f:
            f.write('some text')
        self.port = find_free_tcp_port()
        self.url = 'http://localhost:%s/tests/stuff/demo.txt' % self.port

    def test_main_function(self):

        args = ['nanohttp', '-C', self.module_dir, '-b', str(self.port), ':Static']
        t = threading.Thread(target=main, args=(args,), daemon=True)
        t.start()

        time.sleep(1)

        client = httplib2.Http()
        response, content = client.request(self.url)
        self.assertIsNotNone(response)
        self.assertIsNotNone(content)



import unittest

from nanohttp import Controller, action, context, RestController
from nanohttp.tests.helpers import WsgiAppTestCase


class Users(RestController):
    @action
    def post(self):
        return ', '.join('%s: %s' % (k, v) for k, v in sorted(context.query_string.items(), key=lambda x: x[0]))


class QueryStringTestCase(WsgiAppTestCase):

    class Root(Controller):
        users = Users()

        @action()
        def index(self):
            return ', '.join('%s: %s' % (k, v) for k, v in sorted(context.query_string.items(), key=lambda x: x[0]))

    def test_simple_query_string(self):
        self.assert_get('/?a=1&b=&c=2', expected_response="a: 1, b: , c: 2")
        self.assert_get('/?a=1&b=2', expected_response="a: 1, b: 2")
        self.assert_get('/?a=1&b=2&b=3', expected_response="a: 1, b: ['2', '3']")
        self.assert_get('/', query_string=dict(a=1, b=2), expected_response="a: 1, b: 2")
        self.assert_post('/', query_string=dict(a=1, b=2), expected_response="a: 1, b: 2")

    def test_rest_controller_query_string(self):
        self.assert_post('/users', query_string=dict(a=1, b=2), expected_response="a: 1, b: 2")


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




from .contexts import context


class HttpStatus(Exception):
    status_code = None
    status_text = None
    info = None

    def __init__(self, message=None, reason=None):
        if reason:
            context.response_headers.add_header('X-Reason', reason)
        super().__init__(message or self.status_text)

    @property
    def status(self):
        return '%s %s' % (self.status_code, self)

    def render(self):
        return self.status


class HttpBadRequest(HttpStatus):
    status_code, status_text, info = 400, 'Bad Request', 'Bad request syntax or unsupported method'


class HttpUnauthorized(HttpStatus):
    status_code, status_text, info = 401, 'Unauthorized', 'No permission -- see authorization schemes'


class HttpForbidden(HttpStatus):
    status_code, status_text, info = 403, 'Forbidden', 'Request forbidden -- authorization will not help'


class HttpNotFound(HttpStatus):
    status_code, status_text, info = 404, 'Not Found', 'Nothing matches the given URI'


class HttpMethodNotAllowed(HttpStatus):
    status_code, status_text, info = 405, 'Method Not Allowed', 'Specified method is invalid for this resource'


class HttpConflict(HttpStatus):
    status_code, status_text, info = 409, 'Conflict', 'Request conflict'


class HttpGone(HttpStatus):
    status_code, status_text, info = 410, 'Gone', 'URI no longer exists and has been permanently removed'


class HttpRedirect(HttpStatus):

    def __init__(self, location, *args, **kw):
        context.response_headers.add_header('Location', location)
        super().__init__(*args, **kw)


class HttpMovedPermanently(HttpRedirect):
    status_code, status_text, info = 301, 'Moved Permanently', 'Object moved permanently -- see URI list'


class HttpFound(HttpRedirect):
    status_code, status_text, info = 302, 'Found', 'Object moved temporarily -- see URI list'


# FIXME: Rename to HttpInternalServerError
class InternalServerError(HttpStatus):
    status_code, status_text, info = 500, 'Internal Server Error', 'Server got itself in trouble'

    def __init__(self, exc_info):
        self.exc_info = exc_info

    def render(self):
        from traceback import format_tb
        e_type, e_value, tb = self.exc_info
        return 'Traceback (most recent call last):\n%s\n%s: %s' % (
            ''.join(format_tb(tb)),
            e_type.__name__,
            e_value
        )




import unittest

from nanohttp import Controller, html, HttpBadRequest
from nanohttp.tests.helpers import WsgiAppTestCase


class ExceptionTestCase(WsgiAppTestCase):

    class Root(Controller):
        @html
        def index(self):
            raise HttpBadRequest(reason='blah blah')

    def test_reason(self):
        response, content = self.assert_get('/', status=400)
        self.assertIn('x-reason', response)
        self.assertEqual(response['x-reason'], 'blah blah')


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




import ujson

from nanohttp import Controller, json
from nanohttp.tests.helpers import WsgiAppTestCase


class JsonTestCase(WsgiAppTestCase):

    class Root(Controller):

        @json
        def index(self):
            return {
                'a': 1,
                'b': '2'
            }

        @json
        def via_to_dict(self):
            class Model:
                @staticmethod
                def to_dict():
                    return dict(
                        a=1,
                        b='2'
                    )
            return Model()

        @json
        def error(self):
            class Bad:
                pass
            return Bad()

    def test_json(self):
        resp, content = self.assert_get(
            '/',
            expected_headers={'content-type': 'application/json; charset=utf-8'},
        )
        self.assertDictEqual(ujson.loads(content), {'b': '2', 'a': 1})

        resp, content = self.assert_get('/via_to_dict',)
        self.assertDictEqual(ujson.loads(content), {'b': '2', 'a': 1})

        self.assert_get('/error', status=500)




import unittest
from os.path import join

# noinspection PyProtectedMember
from nanohttp.helpers import load_controller_from_file
from nanohttp.tests.helpers import TEST_DIR


class BootstrapTestCase(unittest.TestCase):

    def test_load_controller_from_file(self):
        controller = load_controller_from_file(':Static')
        self.assertIsNotNone(controller)
        self.assertTrue(hasattr(controller, 'load_app'))

        controller = load_controller_from_file(join(TEST_DIR, 'stuff', 'package_for_test_bootstrapping'))
        self.assertIsNotNone(controller)
        self.assertTrue(hasattr(controller, 'load_app'))

        controller = load_controller_from_file(join(TEST_DIR, 'stuff', 'module_for_test_bootstrapping.py'))
        self.assertIsNotNone(controller)
        self.assertTrue(hasattr(controller, 'load_app'))

        controller = load_controller_from_file(join(TEST_DIR, 'stuff', 'module_for_test_bootstrapping'))
        self.assertIsNotNone(controller)
        self.assertTrue(hasattr(controller, 'load_app'))


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




import threading
import cgi
import wsgiref.util
import wsgiref.headers
from urllib.parse import parse_qs

import ujson

from nanohttp import exceptions
from .helpers import LazyAttribute


class ContextIsNotInitializedError(Exception):
    pass


# FIXME: use __slots__
class Context(object):
    response_encoding = None
    thread_local = threading.local()

    def __init__(self, environ):
        super(Context, self).__init__()
        self.environ = environ
        self.response_headers = wsgiref.headers.Headers()
        self.response_cookies = []

    def __enter__(self):
        self.thread_local.nanohttp_context = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self.thread_local.nanohttp_context

    @LazyAttribute
    def request_content_length(self):
        v = self.environ.get('CONTENT_LENGTH')
        return None if v is None else int(v)

    @LazyAttribute
    def request_content_type(self):
        content_type = self.environ.get('CONTENT_TYPE')
        if content_type:
            return content_type.split(';')[0]
        return None

    @property
    def response_content_type(self):
        content_type = self.response_headers.get('Content-Type')
        if content_type:
            return content_type.split(';')[0]
        return None

    @response_content_type.setter
    def response_content_type(self, v):
        self.response_headers['Content-Type'] = '%s; charset=%s' % (v, self.response_encoding)

    @classmethod
    def get_current(cls):
        if not hasattr(cls.thread_local, 'nanohttp_context'):
            raise ContextIsNotInitializedError("Context is not initialized yet.")
        return cls.thread_local.nanohttp_context

    @LazyAttribute
    def method(self):
        return self.environ['REQUEST_METHOD'].lower()

    @LazyAttribute
    def path(self):
        return self.environ['PATH_INFO']

    @LazyAttribute
    def request_uri(self):
        return wsgiref.util.request_uri(self.environ, include_query=True)

    @LazyAttribute
    def request_scheme(self):
        return wsgiref.util.guess_scheme(self.environ)

    @LazyAttribute
    def query_string(self):
        return {k: v[0] if len(v) == 1 else v for k, v in parse_qs(
            self.environ['QUERY_STRING'],
            keep_blank_values=True,
            strict_parsing=False
        ).items()}

    @LazyAttribute
    def form(self):
        result = {}

        if self.request_content_type == 'application/json':
            fp = self.environ['wsgi.input']
            data = fp.read(self.request_content_length)
            return ujson.decode(data)

        try:
            storage = cgi.FieldStorage(
                fp=self.environ['wsgi.input'],
                environ=self.environ,
                strict_parsing=False,
                keep_blank_values=True
            )
        except TypeError:
            raise exceptions.HttpBadRequest('Cannot parse the request.')

        if storage.list is None or not len(storage.list):
            return {}

        def get_value(f):
            # noinspection PyProtectedMember
            return f.value if isinstance(f,  cgi.MiniFieldStorage) \
                              or (isinstance(f, cgi.FieldStorage) and not f._binary_file) else f

        for k in storage:

            v = storage[k]

            if isinstance(v, list):
                result[k] = [get_value(i) for i in v]
            else:
                result[k] = get_value(v)

        return result

    @LazyAttribute
    def cookies(self):
        if 'HTTP_COOKIE' in self.environ:
            return dict([pair.split('=', 1) for pair in self.environ['HTTP_COOKIE'].split('; ')]) or {}
        else:
            return {}

    def encode_response(self, buffer):
        try:
            if self.response_encoding:
                return buffer.encode(self.response_encoding)
            else:
                return buffer
        except AttributeError:  # pragma: no cover
            raise TypeError('The returned response should has the `encode` attribute, such as `str`.')


class ContextProxy(Context):

    # noinspection PyInitNewSignature
    def __new__(cls) -> Context:
        type_proxy = type('ContextProxy', (object, ), {
            '__getattr__': cls.__getattr__,
            '__setattr__': cls.__setattr__,
            '__delattr__': cls.__delattr__
        })
        # noinspection PyTypeChecker
        return object.__new__(type_proxy)

    def __getattr__(self, key):
        return getattr(Context.get_current(), key)

    def __setattr__(self, key, value):
        setattr(Context.get_current(), key, value)

    def __delattr__(self, key):
        delattr(Context.get_current(), key)


context = ContextProxy()




from nanohttp import Controller, html, settings, context


class Root(Controller):

    @html
    def index(self):  # pragma: no cover
        yield '<html><head><title>nanohttp demo</title></head><body>'
        yield '<h1>nanohttp demo page</h1>'
        yield '<h2>debug flag is: %s</h2>' % settings.debug
        yield '<img src="/static/cat.jpg" />'
        yield '<ul>'
        yield from ('<li><b>%s:</b> %s</li>' % i for i in context.environ.items())
        yield '</ul>'
        yield '</body></html>'
